/*
    Space Invaders arcade machine emulator

    Copyright (c) 1997-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef INVADERS_H_
#define INVADERS_H_

#include <math.h>

#include <emu/emu_standard_machine.h>
#include <emu/emu_memory_iostream.h>
#include <emu/emu_rcfilter.h>
#include <emu/emu_sample_player.h>
#include <cpu/i8080.h>
#include <sound/sn76477.h>
#include <sound/waveform.h>

class T555Astable
{
public:
    T555Astable( double ra, double rb, double c ) {
        setParameters( ra, rb, c );
        active_ = false;
    }

    void setVariableResistor( double ra ) {
        setParameters( ra, rb_, c_ );
    }

    void setParameters( double ra, double rb, double c ) {
        c_ = c;
        ra_ = ra;
        rb_ = rb;

        double t0 = log(3.0) * c * (ra + rb);
        double t1 = log(2.0) * c * rb;
        double t2 = log(2.0) * c * (ra + rb);

        wfg_.setHiLoPeriod( t2, t1, t0 );
    }

    void setAmplitude( int hi, int lo = 0 ) {
        wfg_.setAmplitude( hi, lo );
    }

    bool active() const {
        return active_;
    }

    void setActive( bool active ) {
        if( active_ != active ) {
            if( ! active_ ) {
                wfg_.reset();
            }
            active_ = active;
        }
    }

    void andWithBuffer( int * data, unsigned len, unsigned samplingRate ) {
        if( active_ ) {
            wfg_.andWithBuffer( data, len, samplingRate );
        }
    }

    void setBuffer( int * data, unsigned len, unsigned samplingRate ) {
        if( active_ ) {
            wfg_.setBuffer( data, len, samplingRate );
        }
    }

private:
    double ra_;
    double rb_;
    double c_;
    bool active_;
    TSquareWaveGenerator wfg_;
};

struct SpaceInvadersBoard : public I8080Environment
{
    SpaceInvadersBoard( unsigned char * vram );

    ~SpaceInvadersBoard();

    void run();
    void reset();

    // Implementation of the I8080Environment interface
    unsigned char readByte( unsigned addr );
    void writeByte( unsigned, unsigned char );
    unsigned char readPort( unsigned port );
    void writePort( unsigned, unsigned char );

    // Member variables
    unsigned char   ram_[0x5000];
    unsigned char   port1i_;    // Port 1 in
    unsigned char   port2i_;    // Port 2 in
    unsigned char   port2o_;    // Port 2 out
    unsigned char   port3o_;    // Port 3 out
    unsigned char   port4lo_;   // Port 4 out (lo)
    unsigned char   port4hi_;   // Port 4 out (hi)
    unsigned char   port5o_;    // Port 5 out
    unsigned char * vram_;
    unsigned char   color_prom_[2*1024];
    I8080 *         cpu_;
};

/**
    Space Invaders machine driver.

    @author Alessandro Scotti
    @version 1.1
*/
class SpaceInvadersMachine : public TStandardMachine
{
public:
    virtual ~SpaceInvadersMachine() {
        delete main_board_;
    }

    virtual const TMachineDriverInfo * createDriverInfoInstance() const;

    virtual bool setResourceFile( int id, const unsigned char * buf, unsigned len );

    virtual void run( TFrame * frame, unsigned samplesPerFrame, unsigned samplingRate );

    virtual void reset();

    static TMachine * createInstance() {
        return new SpaceInvadersMachine;
    }

protected:
    SpaceInvadersMachine();

    virtual bool initializeUserInterface( TUserInterface * ui );

    void setActivePalette( unsigned char palette );

    void renderAnalogSounds( TFrame * frame, unsigned samplesPerFrame, unsigned samplingRate );

    void repaintScreen();

private:
    enum {
        ExtendPlayVibratoAmplitude = 70
    };

    unsigned            palette_data_[8][8];
    unsigned char       active_palette_;
    unsigned char       settings_;
    SpaceInvadersBoard *main_board_;
    TSamplePlayer       samples_[10];

    TMixerBuffer            walk_sound_buffer_;
    T555Astable             walk_timer_;
    TRCFilter               walk_rcfilter_1_;
    TRCFilter               walk_rcfilter_2_;

    TMixerBuffer            extend_play_sound_buffer_;
    T555Astable             extend_play_timer_;
    TSquareWaveGenerator    extend_play_tone_;
    unsigned                extend_play_vibrato_offset_;

    SN76477                 ufo_sound_;
};

class SpaceAttackIIMachine : public SpaceInvadersMachine
{
public:
    virtual const TMachineDriverInfo * createDriverInfoInstance() const;

    virtual bool initializeUserInterface( TUserInterface * ui );

    static TMachine * createInstance() {
        return new SpaceAttackIIMachine;
    }

protected:
    SpaceAttackIIMachine();
};

class OzmaWarsMachine : public SpaceInvadersMachine
{
public:
    virtual const TMachineDriverInfo * createDriverInfoInstance() const;

    virtual bool initializeUserInterface( TUserInterface * ui );

    static TMachine * createInstance() {
        return new OzmaWarsMachine;
    }

protected:
    OzmaWarsMachine();
};

class SpaceInvadersPartIIMachine : public SpaceInvadersMachine
{
public:
    virtual const TMachineDriverInfo * createDriverInfoInstance() const;

    virtual bool initializeUserInterface( TUserInterface * ui );

    static TMachine * createInstance() {
        return new SpaceInvadersPartIIMachine;
    }

protected:
    SpaceInvadersPartIIMachine();
};

class LunarRescueMachine : public SpaceInvadersMachine
{
public:
    virtual const TMachineDriverInfo * createDriverInfoInstance() const;

    virtual bool initializeUserInterface( TUserInterface * ui );

    static TMachine * createInstance() {
        return new LunarRescueMachine;
    }

protected:
    LunarRescueMachine();
};

#endif // INVADERS_H_
