/*
    Tickle front-end
    ALSA playback interface

    Copyright (c) 2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "alsa_playback.h"

/*
    Note: this class implements a synchronous interface. ALSA may also invoke a callback to get data
    when needed, here is how it works.

    First, any data that the callback may need must be placed into a single structure:

        struct async_private_data {
            // ...anything
        };

    Then the callback itself is declared as follows:

        static void async_callback( snd_async_handler_t * ahandler )
        {
            snd_pcm_t * playback_handle = snd_async_handler_get_pcm(ahandler);
            struct async_private_data * data = (struct async_private_data *) snd_async_handler_get_callback_private(ahandler);

            // ...do something, probably using snd_pcm_writei() to put data into the buffer
        }

    Finally, the callback function is registered:

        struct async_private_data async_data = { 0 };
        snd_async_handler_t * async_handler;

        snd_async_add_pcm_handler( &async_handler, playback_handle, async_callback, &async_data);

    At this point when the driver invokes the callback whenever it needs more data to play. The minimum amount
    of data that triggers the callback can be configured using snd_pcm_sw_params_set_avail_min().

    Probably, the best implementation of this mechanism inside the AlsaPlayback class would be that of
    always using the same static callback function, passing an AlsaPlayback * as the private data.
    The callback may then just retrieve this pointer and invoke an appropriate virtual method to do
    the real work. Clients would then subclass AlsaPlayback and reimplement this method as needed.
*/

AlsaPlayback::AlsaPlayback( int samplingRate, int samplesPerFrame, const char * device )
{
    num_of_channels_ = 1;
    sampling_rate_ = samplingRate;
    samples_per_frame_ = samplesPerFrame;
    sound_format_ = SND_PCM_FORMAT_S16_LE;
    sound_buffer_ = new signed short [ samplesPerFrame*2 ]; // 16-bit samples

    int result;

    snd_pcm_hw_params_t *hw_params;
    snd_pcm_sw_params_t *sw_params;

    // Open the device for streaming playback
    result = snd_pcm_open ( &playback_handle_, device != 0 ? device : "plughw:0,0", SND_PCM_STREAM_PLAYBACK, 0 );

    // Initialize hardware parameters
    if( result >= 0 ) {
        result = snd_pcm_hw_params_malloc( &hw_params );

        if( result >= 0 ) {
            result = snd_pcm_hw_params_any( playback_handle_, hw_params );
        }

        if( result >= 0 ) {
                result = snd_pcm_hw_params_set_access( playback_handle_, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params_set_format( playback_handle_, hw_params, sound_format_ );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params_set_rate_near( playback_handle_, hw_params, samplingRate, 0 );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params_set_channels(playback_handle_, hw_params, num_of_channels_ );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params_set_buffer_size( playback_handle_, hw_params, samplesPerFrame*3 );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params_set_period_size( playback_handle_, hw_params, samplesPerFrame, 0 );
        }

        if( result >= 0 ) {
            result = snd_pcm_hw_params( playback_handle_, hw_params );
        }

        snd_pcm_hw_params_free( hw_params );
    }
    else {
        playback_handle_ = 0;
    }

    // Initialize software parameters
    if( result >= 0 ) {
        result = snd_pcm_sw_params_malloc( &sw_params );

        if( result >= 0 ) {
            if( result >= 0 ) {
                result = snd_pcm_sw_params_current( playback_handle_, sw_params );
            }

            if( result >= 0 ) {
                result = snd_pcm_sw_params_set_avail_min( playback_handle_, sw_params, samplesPerFrame );
            }

            if( result >= 0 ) {
                result = snd_pcm_sw_params_set_start_threshold( playback_handle_, sw_params, 999999 );
            }

            if( result >= 0 ) {
                result = snd_pcm_sw_params( playback_handle_, sw_params );
            }

            snd_pcm_sw_params_free( sw_params );
        }
    }

    if( (result < 0) && (playback_handle_ != 0) ) {
        snd_pcm_close( playback_handle_ );
        playback_handle_ = 0;
    }

    last_error_ = result;
}

AlsaPlayback::~AlsaPlayback()
{
    if( playback_handle_ != 0 ) {
        snd_pcm_close( playback_handle_ );
    }

    delete [] sound_buffer_;
}

int AlsaPlayback::start()
{
    if( playback_handle_ != 0 ) {
        // Prepare the device
        snd_pcm_prepare( playback_handle_ );

        // Wait until the interface is ready for data (1 second timeout)
        snd_pcm_wait( playback_handle_, 1000 );

        // Prefill buffer
        memset( sound_buffer_, 0, sizeof(short)*samples_per_frame_ );

        for( int i=0; i<2; i++ ) {
            last_error_ = snd_pcm_writei( playback_handle_, sound_buffer_, samples_per_frame_ );

            if( last_error_ < 0 ) {
                break;
            }
        }

        // Start playback
        if( last_error_ >= 0 ) {
            last_error_ = snd_pcm_start( playback_handle_ );
        }
    }

    return last_error_;
}

int AlsaPlayback::stop()
{
    if( playback_handle_ != 0 ) {
        last_error_ = snd_pcm_drop( playback_handle_ );
    }

    return last_error_;
}

int AlsaPlayback::synchAndPlay( const int * buf, int voices )
{
    if( playback_handle_ != 0 ) {
        // Prepare buffer
        signed short * sndBuf = sound_buffer_;

        if( (buf == 0) || (voices <= 0) ) {
            memset( sndBuf, 0, sizeof(short)*samples_per_frame_ );
        }
        else {
            register int mulFactor = 128 / voices;

            for( int i=0; i<samples_per_frame_; i++ ) {
                *sndBuf++ = (*buf++ * mulFactor);
            }
        }

        // Synch with playback
        int frames_to_deliver;

        do {
            frames_to_deliver = snd_pcm_avail_update( playback_handle_ );

            if( frames_to_deliver < 0 ) {
                last_error_ = frames_to_deliver;
                break;
            }
        } while( frames_to_deliver < samples_per_frame_ );

        // Write into the playback buffer
        if( frames_to_deliver >= samples_per_frame_ ) {
            last_error_ = snd_pcm_writei( playback_handle_, sound_buffer_, samples_per_frame_ );
        }
    }

    return last_error_;
}
