/*
    Analog sound emulation library

    Copyright (c) 2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "ase_timer555_astable.h"

ATimer555Astable::ATimer555Astable( AFloat ra, AFloat rb, AFloat c )
{
    c_ = c;
    ra_ = ra;
    rb_ = rb;
    out_hi_ = 3.3;
    out_lo_ = 0.2;
    c_voltage_ = 0; // Voltage across capacitor
    flipflop_ = 1;  // Charging
    reset_ = false;
    control_ = 0;
    
    setVcc( 5.0 );
    setCapacitorCoefficients();
}

void ATimer555Astable::setRa( AFloat ra )
{
    ra_ = ra;
    setCapacitorCoefficients();
}

void ATimer555Astable::setVcc( AFloat vcc )
{
    vcc_ = vcc;

    if( control_ == 0 ) {
        // Assume control is grounded
        threshold_lo_ = (AFloat) (vcc_ / 3);
        threshold_hi_ = threshold_lo_ * 2;
    }
}

void ATimer555Astable::setControl( AChannel * control, AFloat level )
{
    control_ = control;
    control_level_ = level;

    setVcc( vcc_ );
}

void ATimer555Astable::setCapacitorCoefficients()
{
    AFloat d;

    // Charging
    d = c_ * (ra_ + rb_) * ASE::samplingRate;
    b0_ = (AFloat) exp( -1.0 / d );
    a0_ = (AFloat) 1.0 - b0_;

    // Discharging
    d = c_ * rb_ * ASE::samplingRate;
    b1_ = (AFloat) exp( -1.0 / d );
    a1_ = (AFloat) 1.0 - a0_;
}

void ATimer555Astable::setReset( bool reset )
{
    reset_ = reset;

    if( reset ) {
        // Discharge capacitor
        c_voltage_ = 0;
        flipflop_ = 1;
    }
}

void ATimer555Astable::updateBuffer( AFloat * buf, unsigned len, unsigned ofs )
{
    // Update control if defined (regardless of whether we're using it or not)
    if( control_ != 0 ) {
        control_->updateTo( ofs );
    }

    if( reset_ ) {
        // If reset, output is always low
        while( len > 0 ) {
            *buf++ = out_lo_;

            len--;
        }
    }
    else if( control_ == 0 ) {
        // Threshold points are fixed (set by resistors)
        while( len > 0 ) {
            if( flipflop_ ) {
                *buf++ = out_hi_;

                // Capacitor is charging
                c_voltage_ = a0_ * vcc_ + b0_ * c_voltage_;

                if( c_voltage_ >= threshold_hi_ ) {
                    flipflop_ = 0;
                }
            }
            else {
                *buf++ = out_lo_;

                // Capacitor is discharging
                c_voltage_ = b1_ * c_voltage_;

                if( c_voltage_ <= threshold_lo_ ) {
                    flipflop_ = 1;
                }
            }

            len--;
        }
    }
    else {
        // Threshold points are set externally by control voltage
        AFloat * ctl = control_->stream() + streamSize();

        while( len > 0 ) {
            AFloat level = *ctl++ * control_level_;

            if( flipflop_ ) {
                *buf++ = out_hi_;

                // Capacitor is charging
                c_voltage_ = a0_ * vcc_ + b0_ * c_voltage_;

                if( c_voltage_ >= level ) {
                    flipflop_ = 0;
                }
            }
            else {
                *buf++ = out_lo_;

                // Capacitor is discharging
                c_voltage_ = b1_ * c_voltage_;

                if( c_voltage_ <= (level / 2) ) {
                    flipflop_ = 1;
                }
            }

            len--;
        }
    }
}
