/*
    Analog sound emulation library

    Copyright (c) 2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef ASE_TIMER555_ASTABLE_H_
#define ASE_TIMER555_ASTABLE_H_

#include "ase.h"

/*
    555 timer in the typical astable configuration.

    Vcc ->-+---------+--+
           |         |  |
           Ra        |  |
           |     +---+--+--+
           +-----|7  8  4  |
           |     |        3|-->- Vout
           Rb    |  TIMER  | 
           |     |   555   |
           +---+-|6        |  
           |   | |        5|-<-- Vcontrol (*)
           C   +-|2  1     | 
           |     +---+-----+
           |         |
    GND ---+---------+

    (*) if control line not specified, 5 is grounded by default.

*/
class ATimer555Astable : public AChannel
{
public:
    ATimer555Astable( AFloat ra, AFloat rb, AFloat c );

    void setRa( AFloat ra );

    void setVcc( AFloat vcc );

    /**
        Resets the device.

        This is equivalent to pin 4, only that here the function is abstracted
        with a boolean whereas in the real device you have to pull the pin to
        ground to reset. (In fact, to avoid false resets, this pin is usually
        connected directly to Vcc.)

        The device stays in the specified state until this function is called
        again. Note that output is always low in the reset state.

        @param reset true to reset the device, false for normal operation
    */
    void setReset( bool rest );

    void setControl( AChannel * control, AFloat level );

    void setControl( AChannel * control, AFloat r1, AFloat r2 ) {
        setControl( control, r2 / (r1+r2) );
    }

protected:
    virtual void updateBuffer( AFloat * buf, unsigned len, unsigned ofs );

private:
    void setCapacitorCoefficients();

    AFloat ra_;
    AFloat rb_;
    AFloat c_;
    AFloat vcc_;
    AFloat out_hi_;
    AFloat out_lo_;
    AFloat threshold_hi_;
    AFloat threshold_lo_;

    AChannel * control_;
    AFloat control_level_;

    int flipflop_;
    bool reset_;

    AFloat c_voltage_;  // Voltage across capacitor

    AFloat a0_;
    AFloat b0_;
    AFloat a1_;
    AFloat b1_;
};

#endif // ASE_TIMER555_ASTABLE_H_
