/*
    Space Invaders arcade machine emulator

    Copyright (c) 1997-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <math.h>

#include "invaders.h"

enum {
    ScreenWidth     = 224,
    ScreenHeight    = 256,
    ScreenColors    = 8,
    VideoFrequency  = 60,
    CpuCyclesPerInterrupt = 2000000 / (2*VideoFrequency) // 2MHz CPU, 60 FPS, 2 interrupts per frame
};

// External files id (note that declaration order is important here!) and options
enum { 
    EfUfo, EfShot, EfBomb, EfTargetHit, EfStep1, EfStep2, EfStep3, EfStep4, EfUfoHit, EfExtendPlay,
    Ef1, Ef2, Ef3, Ef4, Ef5, Ef6, EfPROM1, EfPROM2,
    OptLives, OptBonus, OptCoinInfo, OptScreenType, OptRedScreen, OptAnalogSounds
};

// Machine info
static TMachineInfo SpaceInvadersInfo = { 
    "invaders", "Space Invaders", "Taito", 1979, ScreenWidth, ScreenHeight, ScreenColors, VideoFrequency
};

static TMachineInfo SpaceInvadersPartIIInfo = { 
    "invadpt2", "Space Invaders Part II", "Taito", 1979, ScreenWidth, ScreenHeight, ScreenColors, VideoFrequency
};

static TMachineInfo LunarRescueInfo = { 
    "lrescue", "Lunar Rescue", "Taito", 1979, ScreenWidth, ScreenHeight, ScreenColors, VideoFrequency
};

static TMachineInfo SpaceAttackIIInfo = { 
    "spaceat2", "Space Attack II", "Zenitone-Microsec LTD", 1980, ScreenWidth, ScreenHeight, ScreenColors, VideoFrequency
};

static TMachineInfo OzmaWarsInfo = { 
    "ozmawars", "Ozma Wars", "SNK", 1979, ScreenWidth, ScreenHeight, ScreenColors, VideoFrequency
};

static TGameRegistrationHandler reg1( &SpaceInvadersInfo,       SpaceInvaders::createInstance );
static TGameRegistrationHandler reg2( &SpaceInvadersPartIIInfo, SpaceInvadersPartII::createInstance );
static TGameRegistrationHandler reg3( &LunarRescueInfo,         LunarRescue::createInstance );
static TGameRegistrationHandler reg4( &SpaceAttackIIInfo,       SpaceAttackII::createInstance );
static TGameRegistrationHandler reg5( &OzmaWarsInfo,            OzmaWars::createInstance );

bool SpaceInvaders::setResourceFile( int id, const unsigned char * buf, unsigned len )
{
    bool result = false;

    if( (id >= EfUfo) && (id <= EfExtendPlay) ) {
        samples_[id-EfUfo].setSample( copySample(buf,len) );
        result = true;
    }
    else {
        result = 0 == resourceHandler()->handle( id, buf, len );
    }

    return result;
}

bool SpaceInvaders::initialize( TMachineDriverInfo * info )
{
    // Declare resources
    resourceHandler()->add( Ef1, "invaders.h", 0x00000000, 0x0800, efROM, main_board_->ram_+0x0000 );
    resourceHandler()->add( Ef2, "invaders.g", 0x00000000, 0x0800, efROM, main_board_->ram_+0x0800 );
    resourceHandler()->add( Ef3, "invaders.f", 0x00000000, 0x0800, efROM, main_board_->ram_+0x1000 );
    resourceHandler()->add( Ef4, "invaders.e", 0x00000000, 0x0800, efROM, main_board_->ram_+0x1800 );

    resourceHandler()->add( EfUfo,         "0.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfShot,        "1.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfBomb,        "2.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfTargetHit,   "3.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfStep1,       "4.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfStep2,       "5.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfStep3,       "6.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfStep4,       "7.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfUfoHit,      "8.wav", 0x00000000, 0x0000, efSample );
    resourceHandler()->add( EfExtendPlay,  "9.wav", 0x00000000, 0x0000, efSample );

    resourceHandler()->addToMachineDriverInfo( info );

    // Setup user interface
    TUserInterface * ui = info->userInterface();
    TUiOption * option;
    TUiOptionGroup * group = ui->addGroup( S_DSW_1, dtDipSwitch );

    option = group->add( OptLives, "Ships per game", 0, 0x03 );
    option->add( "3", 0x00 );
    option->add( "4", 0x01 );
    option->add( "5", 0x02 );
    option->add( "6", 0x03 );

    option = group->add( OptBonus, "Bonus ship at", 1, 0x08 );
    option->add( "1000", 0x08 );
    option->add( "1500", 0x00 );

    option = group->add( OptCoinInfo, "Coin info", 0, 0x80 );
    option->add( S_On, 0x00 );
    option->add( S_Off, 0x80 );

    group = ui->addGroup( "Options", dtDriverOption );

    option = group->add( OptScreenType, "Hardware type", 4, 0x07 );
    option->add( "Monochrome (white)", 0x00 );
    option->add( "Monochrome (cyan)", 0x01 );
    option->add( "Monochrome (green)", 0x02 );
    option->add( "Monochrome (color overlay)", 0x05 );
    option->add( "Color (with PROM)", 0x04 );

    option = group->add( OptRedScreen, "Bomb effect (red screen)", 0, 0x08 );
    option->add( S_Enabled, 0x08 );
    option->add( S_Disabled, 0x00 );

    option = group->add( OptAnalogSounds, "Analog sound emulation", 0, 0x10 );
    option->add( S_Enabled, 0x10 );
    option->add( S_Disabled, 0x00 );

    // Setup the option handler
    optionHandler()->add( &main_board_->port2i_, ui, OptLives, OptBonus, OptCoinInfo );
    optionHandler()->add( &settings_, ui, OptScreenType, OptRedScreen, OptAnalogSounds );

    // Ask notification so internal variables are same as GUI
    return true;
}

SpaceInvaders::SpaceInvaders() :
    walk_timer_( 0, 75000.0, 0.0000001 ),   // RA=variable, RB=75 Kohm, C=0.1 uF
    walk_rcfilter_1_( 100.0, 0.0000047 ),   // R=100 ohm, C=4.7 uF
    walk_rcfilter_2_( 100.0, 0.000010 ),    // R=100 ohm, C=10 uF
    extend_play_timer_( 100000.0, 47000.0, 0.000001 ) // RA=100 Kohm, RB=47 Kohm, C=1 uF
{
    int i;

    createScreen( ScreenWidth, ScreenHeight, ScreenColors );

    // First entry (zero) is black for all palettes
    for( i=0; i<8; i++ ) {
        palette_data_[i][0] = TPalette::encodeColor(0,0,0);
    }

    // Set other colors
    for( i=1; i<8; i++ ) {
        // Palette 0 is black/white
        palette_data_[0][i] = TPalette::encodeColor(255,255,255);

        // Palette 1 is black/light cyan
        palette_data_[1][i] = TPalette::encodeColor(0,169,219);

        // Palette 2 is black/green
        palette_data_[2][i] = TPalette::encodeColor(0,255,0);

        // Palette 3 is black/red (for "all screen red" effect)
        palette_data_[3][i] = TPalette::encodeColor(255,0,0);

        // Palette 4 is for color PROMs
        unsigned r = (i & 0x01) ? 0xFF : 0x00;
        unsigned g = (i & 0x04) ? 0xFF : 0x00;
        unsigned b = (i & 0x02) ? 0xFF : 0x00;
        palette_data_[4][i] = TPalette::encodeColor(r,g,b);

        // Palette 5 is for color overlays (desaturated colors) and is set later
        palette_data_[5][i] = 0;

        // Palette 6 is black/cyan
        palette_data_[6][i] = TPalette::encodeColor(0,255,255);

        // Palette 7 is unused
        palette_data_[7][i] = 0;
    }

    // Set overlay palette
    palette_data_[5][1] = TPalette::encodeColor( 222, 90, 101 );    // Red
    palette_data_[5][3] = TPalette::encodeColor( 255, 252, 146 );   // Yellow
    palette_data_[5][4] = TPalette::encodeColor( 37, 153, 116 );    // Green
    palette_data_[5][7] = TPalette::encodeColor( 240, 240, 240 );   // White

    // Initialize analog circuits
    walk_timer_.setAmplitude( 127, -128 );

    extend_play_timer_.setAmplitude( ~0, 0 ); // On/off (used in "AND" mode)
    extend_play_tone_.setFrequency( 480 ); // 480 Hz square wave
    extend_play_tone_.setAmplitude( 100, -128 + ExtendPlayVibratoAmplitude );

    ufo_sound_.setSLF( 120000, 0.000001 ); // 120K, 1u
    ufo_sound_.setVCO( 3900, 0.00000022, 0, 0 ); // 3.9K, 0.22u (alternate: 8.2K, 0.1u)
    ufo_sound_.setVCO_Select( 1 ); // 0=external, 1=internal
    ufo_sound_.setEnvelopeControl( 100000, 0, 0 ); // 100K, unconnected, unconnected
    ufo_sound_.setAmplifier( 10000, 150000 ); // 10K, 150K
    ufo_sound_.setMixer( snMixer_VCO );
    
    ufo_sound_.setAmplifier( 63 ); // Override schematics values for compatibility with the other sounds

    // Initialize board
    main_board_ = new SpaceInvadersBoard( screen()->bits()->data() );
    active_palette_ = 0xFF;
    settings_ = 0;

    // Create a color PROM for games that do not have one
    memset( main_board_->color_prom_, 7, 0x400 );
    for( i=0; i<32; i++ ) memset( main_board_->color_prom_+i*32+ 2, 4, 7 );
    for( i=0; i<32; i++ ) memset( main_board_->color_prom_+i*32+24, 1, 4 );

    eventHandler()->add( idCoinSlot1,        ptNormal, &main_board_->port1i_, 0x01 );
    eventHandler()->add( idKeyStartPlayer2,  ptNormal, &main_board_->port1i_, 0x02 );
    eventHandler()->add( idKeyStartPlayer1,  ptNormal, &main_board_->port1i_, 0x04 );
    eventHandler()->add( idKeyP1Action1,     ptNormal, &main_board_->port1i_, 0x10 );
    eventHandler()->add( idKeyP2Action1,     ptNormal, &main_board_->port2i_, 0x10 );

    setJoystickHandler( 0, new TJoystickToPortHandler(idJoyP1Joystick1, ptNormal, jm4Way, &main_board_->port1i_, 0x00004020) );
    setJoystickHandler( 1, new TJoystickToPortHandler(idJoyP2Joystick1, ptNormal, jm4Way, &main_board_->port2i_, 0x00004020) );

    // Register this driver
    registerDriver( SpaceInvadersInfo );
}

void SpaceInvaders::setActivePalette( unsigned char index )
{
    if( active_palette_ != index ) {
        active_palette_ = index;
        for( int i=0; i<8; i++ ) {
            palette()->setColor( i, palette_data_[index][i] );
        }
    }
}

void SpaceInvaders::repaintScreen()
{
    // The following code force a refresh of the video RAM, which in
    // turn updates our screen bitmap
    for( unsigned addr=0x2400; addr<0x4000; addr++ ) {
        main_board_->writeByte( addr, main_board_->ram_[addr] );
    }
}

// TODO: design this stuff away when done with experiments

// NOTE: this circuit is a less-than-perfect simulation with hand-coded parameters,
// emulation with values from schematics does not work at this time! :-(

// Noise circuit (shift register configuration is made up but I think few will notice...)
const unsigned LFSR_MASK = 0x30009; 
const int NOISE_EFFECT_DURATION = 633; // Milliseconds
// Noise half period should be about 2.6 (assuming sampling rate is 44100 Hz), 
// but 15 sounds better in "flash"... that circuit needs more analysis anyway
unsigned noise_half_period = 15;
unsigned noise_offset = 0;
unsigned noise_shift_register = 1; // Any non-zero value will do
int noise_duration_; // Sound duration in milliseconds
int noise_value = 0;

RCFilter flash_filter1( 560.0, 0.0000001 ); // 0.1uF
RCFilter flash_filter2( 6800.0, 0.0000001 );

static void update_noise()
{
    if( noise_offset >= noise_half_period ) {
        noise_offset = 0;

        if( noise_shift_register & 1 ) {
            noise_shift_register ^= LFSR_MASK;
            noise_value = 127;
        }
        else {
            noise_value = -128;
        }

        noise_shift_register >>= 1;
    }

    noise_offset++;
}

void SpaceInvaders::run( TFrame * frame, unsigned samplesPerFrame, unsigned samplingRate )
{
    unsigned char old_port3o = main_board_->port3o_;
    unsigned char old_port5o = main_board_->port5o_;

    main_board_->run();

    // If the amplifier is enabled, play the sounds
    // TODO!!!
    if( true || main_board_->port3o_ & 0x20 ) {
        if( settings_ & 0x10 ) {
            renderAnalogSounds( frame, samplesPerFrame, samplingRate );
        }
        else {
            if(  main_board_->port3o_ & 0x01 ) samples_[0].play(pmLooping); else samples_[0].stop();
            if( (main_board_->port3o_ & 0x04) && ((old_port3o & 0x04) == 0) ) samples_[2].play();
            if( (main_board_->port3o_ & 0x10) && ((old_port3o & 0x10) == 0) ) samples_[9].play();
            if( (main_board_->port5o_ & 0x01) && ((old_port5o & 0x01) == 0) ) samples_[4].play();
            if( (main_board_->port5o_ & 0x02) && ((old_port5o & 0x02) == 0) ) samples_[5].play();
            if( (main_board_->port5o_ & 0x04) && ((old_port5o & 0x04) == 0) ) samples_[6].play();
            if( (main_board_->port5o_ & 0x08) && ((old_port5o & 0x08) == 0) ) samples_[7].play();
        }

        if( (main_board_->port5o_ & 0x10) && ((old_port5o & 0x10) == 0) ) samples_[8].play();
        if( (main_board_->port3o_ & 0x02) && ((old_port3o & 0x02) == 0) ) samples_[1].play();
        if( (main_board_->port3o_ & 0x08) && ((old_port3o & 0x08) == 0) ) samples_[3].play();

        // Sample-based sound emulation
        for( int i=0; i<10; i++ ) {
            samples_[i].mix( frame->getMixer(), chMono, samplesPerFrame, samplingRate );
        }
    }

    // Set a minimum number of voices
    frame->getMixer()->getBuffer( chMono, 0, 0 )->setVoices( TMath::max( 4, frame->getMixer()->maxVoicesPerChannel() ) );

    // Set palette and render the video
    if( (settings_ & 0x08) && (main_board_->port3o_ & 0x04) ) {
        setActivePalette( 3 ); // Red screen
    }
    else {
        setActivePalette( settings_ & 0x07 ); // User settings
    }

    frame->setVideo( screen() );
}

void SpaceInvaders::renderAnalogSounds( TFrame * frame, unsigned samplesPerFrame, unsigned samplingRate )
{
#if 0
    // Used as a "simple" VCO for testing...
    static SN76477 sn;

    if( main_board_->port5o_ & 0x10 ) {
        static bool init = true;

        if( init ) {
            init = false;

            // Low-frequency oscillator
            double st1_r1 = 1000000;    // Schmitt trigger R1
            double st1_r2 =  120000;    // Schmitt trigger R2
            double op1_r1 = 1000000;    // Op-amp R1
            double op1_c1 = 0.0000001;  // Op-amp C1

            // If we assume that R1 = 2*R2 (and in the schematics R1=1M and R2=470K,
            // which is close enough) then the output frequency is:
            // (V+ - Vbe) / (2 * R1 * C1 * DVo)
            double v_in = 12.0;
            double v_be =  0.5;
            double d_vo = (st1_r1 + st1_r2) / st1_r2;
            double fq_1 = (v_in - v_be) / (2 * op1_r1 * op1_c1 * d_vo);

            sn.setSLF( fq_1 );

            // Voltage-controlled oscillator
            //sn.setVCO( 10000, 0.00000022, 0, 0 ); // -> 290 Hz
            sn.setVCO( 130.0*2 );  // ...should be up to 1300 Hz, no more...

            // 390 -> 1190

            // Other options
            sn.setVCO_Select( 1 );                  // 0=external, 1=internal
            sn.setEnvelopeControl( 100000, 0, 0 );  // 100K, unconnected, unconnected
            sn.setAmplifier( 10000, 150000 );       // 10K, 150K
            sn.setMixer( snMixer_VCO );

            sn.setAmplifier( 63 ); // Override schematics values for compatibility with the other sounds
        }

        TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );

        sn.enableOutput( true );
        sn.playSound( mixer_buffer->data(), samplesPerFrame, samplingRate );
    }
    else {
        sn.enableOutput( false );
    }
#endif

    // Ufo ship
    if( main_board_->port3o_ & 0x01 ) {
        TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );

        ufo_sound_.enableOutput( true );
        ufo_sound_.playSound( mixer_buffer->data(), samplesPerFrame, samplingRate );
    }
    else {
        ufo_sound_.enableOutput( false );
    }

    // Walk sound circuit
    if( main_board_->port5o_ & 0x0F ) {
        if( ! walk_timer_.active() ) {
            double wr = 0;

            // Load resistors according to port settings
            if( main_board_->port5o_ & 0x01 ) wr +=  1.0 /  40;
            if( main_board_->port5o_ & 0x02 ) wr +=  1.0 /  68;
            if( main_board_->port5o_ & 0x04 ) wr +=  1.0 /  82;
            if( main_board_->port5o_ & 0x08 ) wr +=  1.0 / 100;

            walk_timer_.setVariableResistor( (1/wr)*1000 );
            walk_timer_.setActive( true );
        }

        // Play the sound in a separate buffer
        walk_sound_buffer_.expand( samplesPerFrame );
        walk_timer_.setBuffer( walk_sound_buffer_.data(), samplesPerFrame, samplingRate );
        walk_rcfilter_1_.apply( walk_sound_buffer_.data(), samplesPerFrame, samplingRate );
        walk_rcfilter_2_.apply( walk_sound_buffer_.data(), samplesPerFrame, samplingRate );

        // Mix into the primary buffer
        TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );
        TMixer::mix( mixer_buffer->data(), walk_sound_buffer_.data(), samplesPerFrame );
    }
    else {
        if( walk_timer_.active() ) {
            // Let the sound continue for half a frame (to somewhat compensate for the fact
            // that we are undersampling at 60 Hz while the game runs at 120 Hz), but
            // always run the filters for the whole frame because of the "buffer" effect 
            // introduced by capacitors
            walk_sound_buffer_.expand( samplesPerFrame );
            walk_sound_buffer_.clear();
            walk_timer_.setBuffer( walk_sound_buffer_.data(), samplesPerFrame / 2, samplingRate );
            walk_rcfilter_1_.apply( walk_sound_buffer_.data(), samplesPerFrame, samplingRate );
            walk_rcfilter_2_.apply( walk_sound_buffer_.data(), samplesPerFrame, samplingRate );

            // Mix into the primary buffer
            TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );
            TMixer::mix( mixer_buffer->data(), walk_sound_buffer_.data(), samplesPerFrame );

            // Stop the circuit
            walk_timer_.setActive( false );
            walk_rcfilter_1_.reset();
            walk_rcfilter_2_.reset();
        }
    }

    // Extended play sound circuit
    if( main_board_->port3o_ & 0x10 ) {
        if( ! extend_play_timer_.active() ) {
            extend_play_timer_.setActive( true );
            extend_play_vibrato_offset_ = 0;
        }

        // Play the tone and combine it with the timer
        extend_play_sound_buffer_.expand( samplesPerFrame );
        extend_play_tone_.setBuffer( extend_play_sound_buffer_.data(), samplesPerFrame, samplingRate );
        extend_play_timer_.andWithBuffer( extend_play_sound_buffer_.data(), samplesPerFrame, samplingRate );

        // Add the vibrato (some would say, more precisely, tremolo)
        double extend_play_wave_period = 2 * 3.1415926535 / (samplingRate / 60); // 60 Hz
        int * data = extend_play_sound_buffer_.data();

        for( unsigned n=0; n<samplesPerFrame; n++ ) {
            if( *data <= 50 ) {
                *data += (int) (sin(extend_play_vibrato_offset_*extend_play_wave_period) * ExtendPlayVibratoAmplitude);
            }

            data++;
            extend_play_vibrato_offset_++;
        }

        // Mix it all into the primary buffer
        TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );
        TMixer::mix( mixer_buffer->data(), extend_play_sound_buffer_.data(), samplesPerFrame );
    }
    else {
        extend_play_timer_.setActive( false );
        extend_play_tone_.reset();
    }

    // Flash (base hit) circuit
    if( main_board_->port3o_ & 0x04 ) {
        // If starting the noise, play the feedback effect
        if( noise_duration_ == NOISE_EFFECT_DURATION ) {
            // TODO!!! frame->playForceFeedbackEffect( 0, fxFF_Explosion, 0 );
        }

        TMixerBuffer * mixer_buffer = frame->getMixer()->getBuffer( chMono, samplesPerFrame, 1 );
        int * data_buffer = mixer_buffer->data();

        if( noise_duration_ > 0 ) {
            unsigned d = noise_duration_ * samplingRate / 1000; // Duration in samples

            if( d > samplesPerFrame ) {
                d = samplesPerFrame;
            }

            for( unsigned n=0; n<d; n++ ) {
                update_noise();
                data_buffer[n] += noise_value;
            }

            d = d * 1000 / samplingRate;

            noise_duration_ -= d ? d : noise_duration_;
        }

        flash_filter1.apply( data_buffer, samplesPerFrame, samplingRate );
        flash_filter2.apply( data_buffer, samplesPerFrame, samplingRate );
    }
    else {
        noise_duration_ = NOISE_EFFECT_DURATION; // Duration in milliseconds

        frame->stopForceFeedbackEffect( 0 );
    }
}

void SpaceInvaders::reset()
{
    main_board_->reset();
}

SpaceInvadersBoard::SpaceInvadersBoard( unsigned char * vram )
{
    vram_ = vram;
    cpu_ = new I8080( *this );

    memset( ram_+0x0000, 0, 0x2000 );  // Clear the ROM area
    memset( ram_+0x4000, 0, 0x1000 );  // Clear the ROM area

    port1i_ = 0;
    port2i_ = 0; // DIP switches and player 2 controls
    port2o_ = 0;
    port3o_ = 0;
    port4lo_ = 0;
    port4hi_ = 0;
    port5o_ = 0;

    reset();
}

SpaceInvadersBoard::~SpaceInvadersBoard()
{
    delete cpu_;
}

void SpaceInvadersBoard::run()
{
    cpu_->setCycles( 0 );

    // Before a frame is fully rendered, two interrupts have to occur
    for( int i=0; i<2; i++ ) {
        cpu_->run( CpuCyclesPerInterrupt );

        // Call the proper interrupt (0xD7 is RST 2 and jumps to 0x10,
        // 0xCF is RST 1 and jumps to 0x08)
        cpu_->interrupt( i ? 0xD7 : 0xCF );
    }
}

void SpaceInvadersBoard::reset()
{
    cpu_->reset();

    // Clear the RAM, but avoid the ROM area
    memset( ram_+0x2000, 0, 0x2000 );
    memset( vram_, 0, ScreenWidth*ScreenHeight );
}

unsigned char SpaceInvadersBoard::readByte( unsigned addr )
{
    addr &= 0xFFFF;

    return (addr < sizeof(ram_)) ? ram_[addr] : 0;
}

void SpaceInvadersBoard::writeByte( unsigned addr, unsigned char b )
{
    addr &= 0xFFFF;

    if( (addr >= 0x2000) && addr < 0x4000 ) {
        ram_[addr] = b;

        if( addr >= 0x2400 ) {
            // This is a write to video memory. Since the video screen is rotated, 
            // consecutive bits correspond to vertically consecutive pixels.
            // This is accounted for in the following code, and an extra buffer
            // is used to store the video memory in a more useable form.
		    addr -= 0x2400;

            unsigned x = addr / 32;
            unsigned y = addr % 32;
            unsigned char * vram = vram_ + ((255-y*8)*224) + x;
            unsigned char color = color_prom_[ 0x80 + y + ((x / 8) << 5)] & 0x07;
            for( int i=0; i<8; i++ ) {
                *vram = b & 1 ? color : 0;
			    vram -= ScreenWidth;
                b >>= 1;
            }
        }
    }
}

unsigned char SpaceInvadersBoard::readPort( unsigned port )
{
    unsigned char result = 0;

    switch( port ) {
    case 0:
        result = 0x40;
        break;
    case 1: 
        result = port1i_;
        break;
    case 2: 
        result = port2i_;
        break;
    case 3: 
        result = (unsigned char)(((((unsigned)port4hi_ << 8) | port4lo_) << port2o_) >> 8);
        break;
    }

    return result;
}

void SpaceInvadersBoard::writePort( unsigned addr, unsigned char b )
{
    switch( addr ) {
    case 2:
        port2o_ = b;
        break;
    case 3:
        port3o_ = b;
        break;
    case 4:
        port4lo_ = port4hi_;
        port4hi_ = b;
        break;
    case 5:
        port5o_ = b;
        break;
    case 6:
        // Watchdog timer
        break;
    }
}

bool SpaceInvadersPartII::initialize( TMachineDriverInfo * info )
{
    SpaceInvaders::initialize( info );

    // Update resources
    resourceHandler()->replace( Ef1, "pv.01", 0x00000000 );
    resourceHandler()->replace( Ef2, "pv.02", 0x00000000 );
    resourceHandler()->replace( Ef3, "pv.03", 0x00000000 );
    resourceHandler()->replace( Ef4, "pv.04", 0x00000000 );
    resourceHandler()->add( Ef5, "pv.05", 0x00000000, 0x800, efROM, main_board_->ram_+0x4000 );
    resourceHandler()->add( EfPROM1, "pv06_1.bin", 0x00000000, 0x0400, efColorPROM, main_board_->color_prom_+0x0000 );
    resourceHandler()->add( EfPROM2, "pv07_2.bin", 0x00000000, 0x0400, efColorPROM, main_board_->color_prom_+0x0400 );

    resourceHandler()->addToMachineDriverInfo( info );

    // Update user interface
    TUserInterface * ui = info->userInterface();

    TUiOption * option = ui->replaceOption( OptLives, "Ships per game", 0, 0x03 );
    option->add( "3", 0x00 );
    option->add( "4", 0x01 );
    option->add( S_Unused, 0x02 );
    option->add( S_Unused, 0x03 );

    option = ui->replaceOption( OptBonus, "High score preset mode", 0 , 0x08 );
    option->add( S_Off, 0x00 );
    option->add( S_On, 0x08 );

    ui->removeOption( OptCoinInfo );

    optionHandler()->resynch(ui);

    return true;
}

SpaceInvadersPartII::SpaceInvadersPartII() : SpaceInvaders() 
{
    registerDriver( SpaceInvadersPartIIInfo );
}

bool LunarRescue::initialize( TMachineDriverInfo * info )
{
    SpaceInvaders::initialize( info );

    // Update resources
    resourceHandler()->replace( Ef1, "lrescue.1", 0x00000000 );
    resourceHandler()->replace( Ef2, "lrescue.2", 0x00000000 );
    resourceHandler()->replace( Ef3, "lrescue.3", 0x00000000 );
    resourceHandler()->replace( Ef4, "lrescue.4", 0x00000000 );
    resourceHandler()->add( Ef5, "lrescue.5", 0x00000000, 0x800, efROM, main_board_->ram_+0x4000 );
    resourceHandler()->add( Ef6, "lrescue.6", 0x00000000, 0x800, efROM, main_board_->ram_+0x4800 );
    resourceHandler()->add( EfPROM1, "7643-1.cpu", 0x00000000, 0x0400, efColorPROM, main_board_->color_prom_+0x0000 );

    resourceHandler()->addToMachineDriverInfo( info );

    // Update user interface
    TUserInterface * ui = info->userInterface();

    ui->removeOption( OptBonus );
    ui->removeOption( OptCoinInfo );

    return true;
}

LunarRescue::LunarRescue() : SpaceInvaders() 
{
    registerDriver( LunarRescueInfo );
}

bool SpaceAttackII::initialize( TMachineDriverInfo * info )
{
    SpaceInvaders::initialize( info );

    // Update resources
    resourceHandler()->replace( Ef1, "spaceatt.h", 0x00000000 );
    resourceHandler()->replace( Ef2, "spaceatt.g", 0x00000000 );
    resourceHandler()->replace( Ef3, "spaceatt.f", 0x00000000 );
    resourceHandler()->replace( Ef4, "spaceatt.e", 0x00000000 );

    resourceHandler()->addToMachineDriverInfo( info );

    // Update user interface
    TUserInterface * ui = info->userInterface();

    TUiOption * option = ui->replaceOption( OptCoinInfo, "Coin", 0, 0x80 );
    option->add( S_1Coin1Play, 0x00 );
    option->add( S_2Coin1Play, 0x80 );

    return true;
}

SpaceAttackII::SpaceAttackII() : SpaceInvaders() 
{
    registerDriver( SpaceAttackIIInfo );
}

bool OzmaWars::initialize( TMachineDriverInfo * info )
{
    SpaceInvaders::initialize( info );

    // Update resources
    resourceHandler()->replace( Ef1, "mw01", 0x00000000 );
    resourceHandler()->replace( Ef2, "mw02", 0x00000000 );
    resourceHandler()->replace( Ef3, "mw03", 0x00000000 );
    resourceHandler()->replace( Ef4, "mw04", 0x00000000 );
    resourceHandler()->add( Ef5, "mw05", 0x00000000, 0x800, efROM, main_board_->ram_+0x4000 );
    resourceHandler()->add( Ef6, "mw06", 0x00000000, 0x800, efROM, main_board_->ram_+0x4800 );

    resourceHandler()->addToMachineDriverInfo( info );

    // Update user interface
    TUserInterface * ui = info->userInterface();

    TUiOption * option = ui->replaceOption( OptLives, "Energy", 0, 0x03 );
    option->add( "15000", 0x00 );
    option->add( "20000", 0x01 );
    option->add( "25000", 0x02 );
    option->add( "35000", 0x03 );

    option = ui->replaceOption( OptBonus, "Bonus energy", 0, 0x08 );
    option->add( "15000", 0x00 );
    option->add( "10000", 0x08 );

    option = ui->replaceOption( OptCoinInfo, "Coin", 0, 0x80 );
    option->add( S_1Coin1Play, 0x00 );
    option->add( S_2Coin1Play, 0x80 );

    // Ask notification so internal variables are same as GUI
    return true;
}

OzmaWars::OzmaWars() : SpaceInvaders() 
{
    registerDriver( OzmaWarsInfo );
}
